// This file is part of the program FRYSK.
//
// Copyright 2005, 2008, Red Hat Inc.
//
// FRYSK is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// FRYSK is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with FRYSK.  If not, see <http://www.gnu.org/licenses/>.
// 
// In addition, as a special exception, Red Hat, Inc. gives You the
// additional right to link the code of FRYSK with code not covered
// under the GNU General Public License ("Non-GPL Code") and to
// distribute linked combinations including the two, subject to the
// limitations in this paragraph. Non-GPL Code permitted under this
// exception must only link to the code of FRYSK through those well
// defined interfaces identified in the file named EXCEPTION found in
// the source code files (the "Approved Interfaces"). The files of
// Non-GPL Code may instantiate templates or use macros or inline
// functions from the Approved Interfaces without causing the
// resulting work to be covered by the GNU General Public
// License. Only Red Hat, Inc. may make changes or additions to the
// list of Approved Interfaces. You must obey the GNU General Public
// License in all respects for all of the FRYSK code and other code
// used in conjunction with FRYSK except the Non-GPL Code covered by
// this exception. If you modify this file, you may extend this
// exception to your version of the file, but you are not obligated to
// do so. If you do not wish to provide this exception without
// modification, you must delete this exception statement from your
// version and license this file solely under the GPL without
// exception.

#include <libdwfl.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "jni.hxx"

#include "jnixx/elements.hxx"

using namespace java::lang;

// Assume the method was parameterised with POINTER.
#define DWFL_POINTER ((::Dwfl *)pointer)



jlong
lib::dwfl::Dwfl::dwfl_slap_module_memory(jnixx::env env, jlong module,
					 inua::eio::ByteBuffer memory,
					 jlong address, jlong length) {
  jnixx::jbyteArray bounceBuffer = jnixx::jbyteArray::NewByteArray(env, length);
  ssize_t nread = memory.safeGet(env, (off64_t) address, bounceBuffer, 0, length);
  if (nread < length) {
    bounceBuffer.DeleteLocalRef(env);
    return 0; // what ever
  }
  void **userdatap;
  ::dwfl_module_info((Dwfl_Module*)module, &userdatap, NULL, NULL, NULL, NULL, NULL, NULL);
  (*userdatap) = ::malloc(length);
  memcpy((*userdatap), jbyteArrayElements(env, bounceBuffer).elements(), length);
  bounceBuffer.DeleteLocalRef(env);
  return (jlong)(*userdatap);
}

static int
dwfl_frysk_proc_find_elf(Dwfl_Module *module,
			 void **userdata,
			 const char *module_name, Dwarf_Addr base,
			 char **file_name, Elf **elfp) {	
  // There is an edge case here that was tripped by a corefile case.
  // In that case the specified executable was defined as a relative
  // path (ie ../foo/bar). And that is perfectly valid path name.
  // However when the corefile created its maps it did not convert
  // that path to an absolute path, causing the test below to fail and
  // consider the file ../foo/bar to be an in memory elf image.
  if (module_name[0] == '/') {
    // Return the file name, and then let elfutils decide what to do.
    *file_name = ::strdup(module_name);
    return -1;
  } else {
    // During module creation the Dwfl_Module USERDATA point was set
    // to a raw buffer that contains the entire in-memory section
    // contents.  Create an elf from that.  DwflModule is responsible
    // for freeing the memory.
    char *memory = (char*)(*userdata);
    // Compute the size of the memory using the module's bound.
    ::Dwarf_Addr bound;
    ::dwfl_module_info(module, NULL, NULL, &bound, NULL, NULL, NULL, NULL);
    ssize_t memory_size = bound - base;
    // Turn that memory into an ELF.  The obvious thing to do here is
    // to tell elfutils that it is responsible for the memory - by
    // calling elf_flagelf(ELF_F_MALLOCED) - except all the flags
    // including ELF_F_MALLOCED are not public and setting that
    // particular flag is prohibited.
    *elfp = elf_memory((char*)memory, memory_size);
    if (*elfp == NULL) {
      return -1;
    }
    // Poke something into FILE_NAME so that the caller notices that
    // we've done something to ELF.
    *file_name = ::strdup(module_name);
    return -1;
  }
}



#define DWFL_CALLBACKS ((::Dwfl_Callbacks*)callbacks)

jlong
lib::dwfl::Dwfl::dwfl_callbacks_begin(jnixx::env env, String jdebugInfoPath) {
  jstringUTFChars debugInfoPath = jstringUTFChars(env, jdebugInfoPath);
  char** path = (char**) ::malloc(sizeof (char*));
  if (path == NULL) {
    return 0;
  }
  *path = ::strdup(debugInfoPath.elements());
  jlong callbacks = (jlong) ::malloc(sizeof(Dwfl_Callbacks));
  ::memset(DWFL_CALLBACKS, 0, sizeof(Dwfl_Callbacks));
  DWFL_CALLBACKS->find_elf = &::dwfl_frysk_proc_find_elf;
  DWFL_CALLBACKS->find_debuginfo = &::dwfl_standard_find_debuginfo;
  DWFL_CALLBACKS->debuginfo_path = path;
  return callbacks;
}

void
lib::dwfl::Dwfl::dwfl_callbacks_end(jnixx::env env, jlong callbacks) {
  ::free(*DWFL_CALLBACKS->debuginfo_path);
  ::free(DWFL_CALLBACKS->debuginfo_path);
  ::free(DWFL_CALLBACKS);
}

jlong
lib::dwfl::Dwfl::dwfl_begin(jnixx::env env, jlong callbacks) {
  return (jlong) ::dwfl_begin(DWFL_CALLBACKS);
}

void
lib::dwfl::Dwfl::dwfl_end(jnixx::env env, jlong pointer) {
  ::dwfl_end(DWFL_POINTER);
}

void
lib::dwfl::Dwfl::dwfl_report_begin(jnixx::env env, jlong pointer) {
  ::dwfl_report_begin(DWFL_POINTER);
}

void
lib::dwfl::Dwfl::dwfl_report_end(jnixx::env env, jlong pointer) {
  ::dwfl_report_end(DWFL_POINTER, NULL, NULL);
}


jlong
lib::dwfl::Dwfl::dwfl_report_module(jnixx::env env, jlong pointer,
				    String jmoduleName, jlong low, jlong high) {
  jstringUTFChars moduleName = jstringUTFChars(env, jmoduleName);
  Dwfl_Module *module = ::dwfl_report_module(DWFL_POINTER,
					     moduleName.elements(),
					     (::Dwarf_Addr) low,
					     (::Dwarf_Addr) high);
  return (jlong)module;
}
