// This file is part of the program FRYSK.
//
// Copyright 2007, 2008, Red Hat Inc.
// Copyright 2016, Andrew Cagney
//
// FRYSK is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// FRYSK is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with FRYSK.  If not, see <http://www.gnu.org/licenses/>.
//
// In addition, as a special exception, Red Hat, Inc. gives You the
// additional right to link the code of FRYSK with code not covered
// under the GNU General Public License ("Non-GPL Code") and to
// distribute linked combinations including the two, subject to the
// limitations in this paragraph. Non-GPL Code permitted under this
// exception must only link to the code of FRYSK through those well
// defined interfaces identified in the file named EXCEPTION found in
// the source code files (the "Approved Interfaces"). The files of
// Non-GPL Code may instantiate templates or use macros or inline
// functions from the Approved Interfaces without causing the
// resulting work to be covered by the GNU General Public
// License. Only Red Hat, Inc. may make changes or additions to the
// list of Approved Interfaces. You must obey the GNU General Public
// License in all respects for all of the FRYSK code and other code
// used in conjunction with FRYSK except the Non-GPL Code covered by
// this exception. If you modify this file, you may extend this
// exception to your version of the file, but you are not obligated to
// do so. If you do not wish to provide this exception without
// modification, you must delete this exception statement from your
// version and license this file solely under the GPL without
// exception.

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include <libdwfl.h>
#include LIBUNWIND_TARGET_H

#include <libelf.h>
#include <gelf.h>

#include "jni.hxx"

#include "jnixx/exceptions.hxx"
#include "jnixx/logging.hxx"
#include "jnixx/elements.hxx"
#include "jnixx/bounds.hxx"

using namespace java::lang;
using namespace lib::unwind;

#define CURSOR_MAGIC 0xfab

struct jni_cursor {
  unw_cursor_t cursor;
  jobject addressSpace;
  int magic;
};

#define UNW_PROC_INFO ((unw_proc_info_t*)unwProcInfo)
#define UNW_ADDRESS_SPACE ((unw_addr_space_t)unwAddressSpace)
#define UNW_CURSOR (&((jni_cursor*)unwCursor)->cursor)
#define FINE env, (GetFine(env))

#ifndef MAX_VDSO_SIZE
# define MAX_VDSO_SIZE ((size_t) sysconf (_SC_PAGESIZE))
#endif

/**
 * The address space is going to be accessed in different JNI contexts
 * to when it is passed to libunwind.  Consequently, need to get the
 * ENV from JNI and can't rely on it being passed in.
 */
#define UNW_CONTEXT \
  jnixx::env env = Object::_env_();			\
  AddressSpace addressSpace = AddressSpace(env, (jobject)contextArg)
    

/*
 * Callback: Get misc. proc info
 */
static int
find_proc_info(::unw_addr_space_t as, ::unw_word_t ip,
	       ::unw_proc_info_t *pip, int need_unwind_info,
	       void *contextArg) {
  UNW_CONTEXT;
  ProcInfo procInfo = ProcInfo::New(env,
				    addressSpace.GetUnwinder(env),
				    (long) ::malloc(sizeof(unw_proc_info_t)));
  int ok = addressSpace.findProcInfo(env, (jlong)ip,
				     (jboolean)need_unwind_info,
				     procInfo);
  if (ok < 0)
    return ok;
  // Extract the info.
  memcpy(pip, (void*) procInfo.GetUnwProcInfo(env), sizeof (unw_proc_info_t));
  procInfo.DeleteLocalRef(env);
  return 0;
}

/*
 * Callback: Free space allocated during find_proc_info
 */
static void
put_unwind_info(::unw_addr_space_t as, ::unw_proc_info_t *proc_info,
		void *contextArg) {
  UNW_CONTEXT;
  // This is passing up a stack pointer, which may then be freed.
  ProcInfo procInfo = ProcInfo::New(env, addressSpace.GetUnwinder(env),
				    (jlong) proc_info);
  addressSpace.putUnwindInfo(env, procInfo);
}

/*
 * Callback: Get the head of the dynamic unwind registration list.
 * There is never any dynamic info in our case.
 */
static int
get_dyn_info_list_addr(::unw_addr_space_t as, ::unw_word_t *dilap,
		       void *contextArg) {
  return -UNW_ENOINFO;
}

/*
 * Callback: Perform memory read/write.  Implement as copy-in,
 * copy-out.
 */
static int
access_mem(::unw_addr_space_t as, ::unw_word_t addr,
	   ::unw_word_t *valp, int write, void *contextArg) {
  // fprintf(stderr, "access_mem as: %p addr: %llx valp: %p write: %d context: %p\n", as, (long long) addr, valp, write, contextArg);
  UNW_CONTEXT;
  try {
    jnixx::jbyteArray jtmp
      = jnixx::jbyteArray::NewByteArray(env, sizeof(unw_word_t));
    // fprintf(stderr, "access_mem have jtmp\n");
    jbyteArrayElements tmp = jbyteArrayElements(env, jtmp);
    // fprintf(stderr, "access_mem have tmp\n");
    {
      jbyte *elements = tmp.elements();
      // fprintf(stderr, "access_mem %llx copy-in %p to %p\n", (long long) addr, valp, elements);
      memcpy(elements, valp, sizeof(unw_word_t));
      tmp.release();
    }
    int ret = addressSpace.accessMem(env, (jlong) addr,
				     jtmp, (jboolean) write);
    {
      jbyte *elements = tmp.elements();
      // fprintf(stderr, "access_mem %llx copy-out %p to %p\n", (long long) addr, elements, valp);
      memcpy(valp, elements, sizeof(unw_word_t));
      tmp.release();
    }
    jtmp.DeleteLocalRef(env);
    return ret;
  } catch (Throwable t) {
    if (t.IsInstanceOf(env, frysk::UserException::_class_(env))) {
      // We have to catch all RuntimeExceptions here since there is no
      // indicator for just "invalid memory location".  Core files
      // might have "holes" in their memory.
      return -UNW_EINVAL;
    } else {
      throw t;
    }
  }
}

/*
 * Callback: perform register read/write
 */
static int
access_reg(::unw_addr_space_t as, ::unw_regnum_t regnum,
	   ::unw_word_t *valp, int write, void *contextArg) {
  UNW_CONTEXT;
  jnixx::jbyteArray jtmp
    = jnixx::jbyteArray::NewByteArray(env, sizeof(unw_word_t));
  jbyteArrayElements tmp = jbyteArrayElements(env, jtmp);
  // Map the REGNUM back to the published ENUM.
  Number num = TARGET_REGISTERS::valueOf(env, regnum);
  ::memcpy(tmp.elements(), valp, sizeof(unw_word_t));
  tmp.release();
  if (write)
    addressSpace.setReg(env, num, *valp);
  else
    *valp = addressSpace.getReg(env, num);
  num.DeleteLocalRef(env);
  jtmp.DeleteLocalRef(env);
  return 0;
}

/*
 * Callback: Perform a floating point register read/write
 */
static int
access_fpreg(::unw_addr_space_t as, ::unw_regnum_t regnum,
	     ::unw_fpreg_t *fpvalp, int write, void *contextArg) {
  UNW_CONTEXT;
  jnixx::jbyteArray jtmp
    = jnixx::jbyteArray::NewByteArray(env, sizeof (unw_fpreg_t));
  jbyteArrayElements tmp = jbyteArrayElements(env, jtmp);
  // Map the REGNUM back to the published ENUM.
  Number num = TARGET_REGISTERS::valueOf(env, regnum);
  // Implement read/modify/write style op.
  ::memcpy(tmp.elements(), fpvalp, sizeof(unw_fpreg_t));
  tmp.release();
  int ret = addressSpace.accessReg(env, num, jtmp, (jboolean) write);
  ::memcpy(fpvalp, tmp.elements(), sizeof(unw_fpreg_t));
  tmp.release();
  num.DeleteLocalRef(env);
  jtmp.DeleteLocalRef(env);
  return ret;
}

/*
 * Callback: Resumes the process at the provided stack level.  We
 * never resume a process through libunwind.
 */
static int
resume(::unw_addr_space_t as, ::unw_cursor_t *cp, void *contextArg) {
  return -UNW_EINVAL;
}

/*
 * Callback: Returns the name of the procedure that the provided
 * address is in as well as the offset from the start of the
 * procedure.
 */
static int
get_proc_name(::unw_addr_space_t as,
	      ::unw_word_t addr, char *bufp,
	      size_t buf_len, ::unw_word_t *offp, void *contextArg) {
  // This should never be called, always return an error.
  return -UNW_ENOMEM;
}

jlong
TARGET::createCursor(jnixx::env env,
		     AddressSpace addressSpace,
		     jlong unwAddressSpace) {
  jni_cursor *cursor = new(jni_cursor);
  logf(FINE, "createCursor %p using address-space %p arg %llx",
       cursor, UNW_ADDRESS_SPACE, (long long) addressSpace._object);
  // XXX: Need to zero out the cursor, as unw_init_remote doesn't seem
  // to do it.
  memset(cursor, 0, sizeof(jni_cursor));
  // Make global as object is used across JNI calls
  cursor->magic = CURSOR_MAGIC;
  cursor->addressSpace = env.NewGlobalRef(addressSpace._object);
  unw_init_remote(&cursor->cursor, UNW_ADDRESS_SPACE, cursor->addressSpace);
  logf(FINE, "createCursor %p global address-space ref %llx",
       cursor, (long long) cursor->addressSpace);
  return (jlong)cursor;
}

void
TARGET::destroyCursor(jnixx::env env, jlong unwCursor) {
  jni_cursor *cursor = (jni_cursor*)unwCursor;
  logf(FINE, "destroyCursor %p", cursor);
  if (cursor->magic != CURSOR_MAGIC) {
    fprintf(stderr, "\n%s: cursor %p has bad CURSOR_MAGIC number\n",
	    __func__, cursor);
    abort();
  }
  logf(FINE, "destroyCursor %p deleting global address-space ref %llx",
       cursor, (long long) cursor->addressSpace);
  env.DeleteGlobalRef(cursor->addressSpace);
  ::free(cursor);
}

jlong
TARGET::copyCursor(jnixx::env env, jlong unwCursor) {
  jni_cursor *original = (jni_cursor*) unwCursor;
  logf(FINE, "copyCursor %p", original);
  if (original->magic != CURSOR_MAGIC) {
    fprintf(stderr, "\n%s: cursor %p has bad CURSOR_MAGIC number\n",
	    __func__, original);
    abort();
  }
  jni_cursor *copy = new(jni_cursor);
  // Create a local copy of the unwind cursor
  ::memcpy(&copy->cursor, &original->cursor, sizeof(original->cursor));
  copy->magic = CURSOR_MAGIC;
  copy->addressSpace = env.NewGlobalRef(original->addressSpace);
  logf(FINE, "copyCursor %p to %p global address-space ref %llx",
       original, copy, (long long) copy->addressSpace);
  return (jlong) copy;
}

jlong
TARGET::createAddressSpace(jnixx::env env, ByteOrder byteOrder) {
  logf(FINE, "createAddressSpace, byteOrder %d",
       (int) byteOrder.hashCode(env));
  static unw_accessors_t accessors = {
    find_proc_info ,
    put_unwind_info,
    get_dyn_info_list_addr,
    access_mem,
    access_reg,
    access_fpreg,
    resume,
    get_proc_name
  };
  unw_addr_space_t as = unw_create_addr_space(&accessors, (int) byteOrder.hashCode(env));
  unw_set_caching_policy(as, UNW_CACHE_NONE);
  logf(FINE, "createAddressSpace at %p", as);
 return (jlong) as;
}

void
TARGET::destroyAddressSpace(jnixx::env env, jlong unwAddressSpace) {
  logf(FINE, "destroyAddressSpace %lx", (long)UNW_ADDRESS_SPACE);
  unw_destroy_addr_space(UNW_ADDRESS_SPACE);
}

void
TARGET::setCachingPolicy(jnixx::env env, jlong unwAddressSpace,
			 CachingPolicy cachingPolicy) {
  log(FINE, "setCachingPolicy, cachingPolicy:", cachingPolicy);
  unw_set_caching_policy(UNW_ADDRESS_SPACE,
                         (unw_caching_policy_t) cachingPolicy.hashCode(env));
}

jint
TARGET::isSignalFrame(jnixx::env env, jlong unwCursor) {
  logf(FINE, "isSignalFrame: %p", UNW_CURSOR);
  return unw_is_signal_frame(UNW_CURSOR);
}

jint
TARGET::step(jnixx::env env, jlong unwCursor) {
  logf(FINE, "step cursor: %p", UNW_CURSOR);
  return unw_step(UNW_CURSOR);
}

static void
verifyBounds(jnixx::env env, jlong offset, jint length,
	     jnixx::jbyteArray bytes, jint start, int size) {
  verifyBounds(env, bytes, start, length);
  if (offset < 0)
    ArrayIndexOutOfBoundsException::New(env, offset).Throw(env);
  if (offset + length > size)
    ArrayIndexOutOfBoundsException::New(env, offset + length).Throw(env);
}

void
TARGET::getRegister(jnixx::env env, jlong unwCursor,
		    Number num, jlong offset, jint length,
		    jnixx::jbyteArray jbytes, jint start) {
  int regNum = num.intValue(env);
  logf(FINE, "getRegister: %d from cursor: %p offset: %ld length: %d start: %d",
       regNum, UNW_CURSOR, (long) offset, (int)length, (int)start);
  int status;
  union {
    unw_word_t w;
    unw_fpreg_t fp;
  } word;
  int size;
  if (unw_is_fpreg(regNum))
    size = sizeof(word.fp);
  else
    size = sizeof(word.w);
  verifyBounds(env, offset, length, jbytes, start, size);
  if (unw_is_fpreg(regNum)) {
    status = unw_get_fpreg(UNW_CURSOR,
			   (::unw_regnum_t) regNum,
			   &word.fp);
  } else {
    status = unw_get_reg(UNW_CURSOR,
			 (::unw_regnum_t) regNum,
			 &word.w);
    logf(FINE, "getRegister status %d %lx", status, (long)word.w);
  }
  if (status != 0)
    RuntimeException::ThrowNew(env, "get register failed");
  jbyteArrayElements bytes = jbyteArrayElements(env, jbytes);
  memcpy(bytes.elements() + start, (uint8_t*)&word + offset, length);
}

void
TARGET::setRegister(jnixx::env env, jlong unwCursor,
		    Number num,
		    jlong offset, jint length,
		    jnixx::jbyteArray jbytes, jint start) {
  int regNum = num.intValue(env);
  int status;
  union {
    unw_word_t w;
    unw_fpreg_t fp;
  } word;
  int size;
  if (unw_is_fpreg(regNum))
    size = sizeof(word.fp);
  else
    size = sizeof(word.w);
  verifyBounds(env, offset, length, jbytes, start, size);
  if (unw_is_fpreg(regNum))
    status = unw_get_fpreg(UNW_CURSOR, (::unw_regnum_t) regNum, &word.fp);
  else
    status = unw_get_reg(UNW_CURSOR, (::unw_regnum_t) regNum, &word.w);
  if (status != 0)
    RuntimeException::ThrowNew(env, "set register failed");
  jbyteArrayElements bytes = jbyteArrayElements(env, jbytes);
  memcpy((uint8_t*)&word + offset, bytes.elements() + start, length);
  bytes.release();
  if (unw_is_fpreg(regNum))
    status = unw_set_fpreg(UNW_CURSOR, regNum, word.fp);
  else
    status = unw_set_reg(UNW_CURSOR, regNum, word.w);
  if (status != 0)
    RuntimeException::ThrowNew(env, "set register failed");
}

jlong
TARGET::getSP(jnixx::env env, jlong unwCursor) {
  unw_word_t sp;
  int status = unw_get_reg(UNW_CURSOR, UNW_REG_SP, &sp);
  if (status < 0)
    return 0; // bottom of stack.
  else
    return sp;
}

jlong
TARGET::getIP(jnixx::env env, jlong unwCursor) {
  unw_word_t ip;
  int status = unw_get_reg(UNW_CURSOR, UNW_REG_IP, &ip);
  if (status < 0)
    return 0; // bottom of stack.
  else
    return ip;
}

jlong
TARGET::getCFA(jnixx::env env, jlong unwCursor) {
#ifdef UNW_TARGET_X86
#define FRYSK_UNW_REG_CFA UNW_X86_CFA
#else
#ifdef UNW_TARGET_X86_64
#define FRYSK_UNW_REG_CFA UNW_X86_64_CFA
#else
  // This is wasteful, but there is no generic UNW_REG_CFA.
  // So just unwind and return the stack pointer.
#define FRYSK_UNW_REG_CFA UNW_REG_SP
  // Need to copy everything including magic and the addressSpace
  // handle.
  jni_cursor copy;
  memcpy(&copy, UNW_CURSOR, sizeof (copy));
  if (unw_step(&copy.cursor) < 0)
    return 0;
  unwCursor = (jlong) &copy;
#endif
#endif

  unw_word_t cfa;
  int status = unw_get_reg(UNW_CURSOR, FRYSK_UNW_REG_CFA, &cfa);
  if (status < 0)
    return 0; // bottom of stack.
  else
    return cfa;
}


jint
TARGET::getContext(jnixx::env env, jlong context) {
  return (jint) unw_getcontext((::unw_context_t *) context);
}

jlong
TARGET::getProcInfo(jnixx::env env, jlong unwCursor) {
  logf(FINE, "getProcInfo cursor: %lx", (long) UNW_CURSOR);
  unw_proc_info_t *procInfo
    = (::unw_proc_info_t *) ::malloc(sizeof (::unw_proc_info_t));
  int ret = unw_get_proc_info(UNW_CURSOR, procInfo);

  logf(FINE, "getProcInfo finished get_proc_info %lx", (long) procInfo);
  if (ret < 0) {
    ::free(procInfo);
    return 0;
  } else {
    return (jlong)procInfo;
  }
}

void
TARGET::destroyProcInfo(jnixx::env env, jlong unwProcInfo) {
  ::free(UNW_PROC_INFO);
}

/**
 * The following are local memory image address space memory accessors
 * used by unw_get_unwind_table and run_cfi_program to access and
 * interpret the eh_frame_hdr.  The BYTES pointer contains a SIZE byte
 * memory mapped image.  BASE, BOUND, and OFFSET define how to map
 * accesses to that address.  While perhaps not the most efficient,
 * this should help with debugging.
 */
#define IMAGE_MAGIC 0xfeed
struct image {
  int magic;
  char *name;
  char *bytes;
  size_t size;
  unw_word_t base;
  unw_word_t bound;
  unw_word_t offset;
  image(const char *name, char* bytes, size_t size, unw_word_t base, unw_word_t bound, unw_word_t offset) {
    this->magic = IMAGE_MAGIC;
    this->name = strdup(name);
    this->bytes = bytes;
    this->size = size;
    this->base = base;
    this->bound = bound;
    this->offset = offset;
  }
};

static int 
image_access_mem(unw_addr_space_t as, unw_word_t addr,
		 unw_word_t *val, int write, void *arg) {
  struct image *image = (struct image*) arg;
  if (image->magic != IMAGE_MAGIC) {
    fprintf(stderr, "%s: image %p has bad IMAGE_MAGIC number\n",
	    __func__, image);
    abort();
  }
  // Writing is not supported
  if (write)
    return -UNW_EINVAL;
  if (addr < image->base || addr > image->bound - sizeof(unw_word_t)) {
    fprintf(stderr, "%s: addr %llx out of bounds [%llx..%llx)\n", __func__,
	    (long long) addr, (long long) image->base, (long long) image->bound);
    return -UNW_EINVAL;
  }
  unw_word_t offset = (addr - image->base) + image->offset;
  if (offset < 0 || offset > image->size - sizeof(unw_word_t)) {
    fprintf(stderr, "%s: offset: %llx out of bounds [0..%llx)\n", __func__,
	    (long long) offset, (long long) image->size);
    return -UNW_EINVAL;
  }
  if (offset & (sizeof(unw_word_t) - 1)) {
    fprintf(stderr, "%s: offset: %llx not aligned to word boundary %zx\n", __func__,
	    (long long) offset, sizeof(unw_word_t) - 1);
    return -UNW_EINVAL;
  }
  // fprintf(stderr, "%s: %s addr: %llx offset: %llx\n", __func__, image->name, (long long) addr, (long long) offset);
  *val = *(unw_word_t *) (image->bytes + offset);
  return UNW_ESUCCESS;
}

static void
image_put_unwind_info(::unw_addr_space_t as, ::unw_proc_info_t *proc_info,
		      void *arg) {
  struct image *image = (struct image*) arg;
  if (image->magic != IMAGE_MAGIC) {
    fprintf(stderr, "%s: bad magic number\n", __func__);
    return;
  }
  munmap(image->bytes, image->size);
  ::free(image->name);
  ::free(image);
}

static unw_accessors_t image_accessors = {
  NULL, image_put_unwind_info, NULL, image_access_mem, NULL, NULL, NULL, NULL
};

// Return NULL when eh_frame_hdr cannot be found.
// Also fills in ip->start_ip, ip->end_ip and ip->gp.
// peh_vaddr will point to the address of the eh_frame_hdr in the main
// address space of the inferior.
static struct image *
get_eh_frame_hdr_addr(jnixx::env env, frysk::rsl::Log fine, const char *name,
		      unw_proc_info_t *pi,
		      char *bytes,  size_t size,
		      unw_word_t segbase, unw_word_t segbound,
		      unw_word_t segoffset, unw_word_t *peh_vaddr)
{
  if (elf_version(EV_CURRENT) == EV_NONE)
    return NULL;

  Elf *elf = elf_memory(bytes, size);
  if (elf == NULL)
    return NULL;

  GElf_Ehdr ehdr;
  if (gelf_getehdr(elf, &ehdr) == NULL)
    return NULL;
  
  GElf_Phdr phdr;
  int ptxt_ndx = -1, peh_hdr_ndx = -1, pdyn_ndx = -1;
  for (int i = 0; i < ehdr.e_phnum; i++) {
      if (gelf_getphdr (elf, i, &phdr) == NULL)
	return NULL;
      
      switch (phdr.p_type) {
        case PT_LOAD:
	  if ((long long)segoffset == (long long) -1) {
	    // I suspect this was to handle the old pre-link hack?
	    if (phdr.p_vaddr == segbase)
	      ptxt_ndx = i;
	  } else {
	    if (phdr.p_offset == segoffset)
	      ptxt_ndx = i;
	  }
          break;
	  
        case PT_GNU_EH_FRAME:
          peh_hdr_ndx = i;
          break;
	  
        case PT_DYNAMIC:
          pdyn_ndx = i;
          break;

        default:
          break;
        }
    }

  Elf_Data *debug_frame_data = NULL;
  size_t shstrndx;
  if (elf_getshdrstrndx (elf, &shstrndx) >= 0) {
      Elf_Scn *scn = NULL;
      while ((scn = elf_nextscn (elf, scn)) != NULL
	     && debug_frame_data == NULL) {
	  GElf_Shdr shdr;
	  if (gelf_getshdr (scn, &shdr) != NULL
	      && shdr.sh_type == SHT_PROGBITS) {
	      const char *name = elf_strptr (elf, shstrndx, shdr.sh_name);
	      if (strcmp (name, ".debug_frame") == 0)
		debug_frame_data = elf_getdata (scn, NULL);
	    }
	}
    }

  if (ptxt_ndx == -1 || (peh_hdr_ndx == -1 && debug_frame_data == NULL))
    return NULL;

  GElf_Phdr ptxt, peh_hdr;
  if (gelf_getphdr (elf, ptxt_ndx, &ptxt) == NULL)
    return NULL;

  if (gelf_getphdr (elf, peh_hdr_ndx, &peh_hdr) == NULL)
    return NULL;

  /* For dynamicly linked executables and shared libraries, DT_PLTGOT
     is the value that data-relative addresses are relative to for
     that object.  We call this the "gp". */
  unw_word_t gp = 0;
  if (pdyn_ndx != -1) {
      GElf_Phdr pdyn;
      if (gelf_getphdr (elf, pdyn_ndx, &pdyn) == NULL)
	return NULL;

      Elf_Scn *pdyn_scn = gelf_offscn(elf, pdyn.p_offset);
      if (pdyn_scn == NULL)
	return NULL;

      Elf_Data *pdyn_data;
      pdyn_data = elf_getdata (pdyn_scn, NULL);
      if (pdyn_data == NULL)
	return NULL;

      GElf_Shdr pdyn_shdr;
      if (gelf_getshdr (pdyn_scn, &pdyn_shdr) == NULL)
	return NULL;

      for (unsigned int i = 0;
	   i < pdyn_shdr.sh_size / pdyn_shdr.sh_entsize; i++) {
	  GElf_Dyn dyn;
	  if (gelf_getdyn (pdyn_data, i, &dyn) == NULL)
	    return NULL;

	  if (dyn.d_tag == DT_PLTGOT) {
	      /* Assume that _DYNAMIC is writable and GLIBC has
		 relocated it (true for x86 at least). */
	      gp = dyn.d_un.d_ptr;
	      break;
	    }
	}
  } else {
    /* Otherwise this is a static executable with no _DYNAMIC.  Assume
       that data-relative addresses are relative to 0, i.e.,
       absolute. */
    gp = 0;
  }

  pi->gp = gp;
  pi->start_ip = segbase;
  if (segbound > segbase + ptxt.p_memsz) {
    segbound = segbase + ptxt.p_memsz;
    logf(env, fine, "adjusted segbound to %llx using p_memsz", (long long)segbound);
  }
  pi->end_ip = segbound;
  pi->format = UNW_INFO_FORMAT_REMOTE_TABLE;
  pi->flags = 0;

  if (debug_frame_data != NULL && debug_frame_data->d_buf != NULL
      && debug_frame_data->d_size != 0) {
    pi->flags |= UNW_PI_FLAG_DEBUG_FRAME;
    *peh_vaddr = 0;
    logf(env, fine, "frame table at %llx", (long long) *peh_vaddr);
    return new image(name, bytes, size, 0, size, 0);
  } else {
    // There was code setting this to just "peh_hdr.p_vaddr", I
    // suspect that is pre-link related.
    *peh_vaddr = segbase + peh_hdr.p_vaddr;
    logf(env, fine, "unwind table at %llx", (long long) *peh_vaddr);
    return new image(name, bytes, size, segbase, segbound, segoffset);
  }
}

static jint
fillProcInfoFromImage(jnixx::env env, frysk::rsl::Log fine,
		      const char* name,
		      unw_proc_info_t* pi,
		      jboolean needUnwindInfo,
		      void *bytes, size_t size, unw_word_t ip,
		      unw_word_t segbase, unw_word_t segbound, unw_word_t segoffset) {
  logf(env, fine,
       "fillProcInfoFromImage %s pi: %p bytes: %p size: %zd ip: %llx segbase: %llx segbound: %llx segoffset: %llx",
       name, pi, bytes, size, (long long) ip,
       (long long) segbase, (long long) segbound, (long long) segoffset);
  
  unw_word_t peh_vaddr;
  struct image *image = get_eh_frame_hdr_addr(env, fine, name, pi, (char*)bytes, size,
					      segbase, segbound, segoffset,
					      &peh_vaddr);
  if (image == NULL) {
    logf(env, fine, "get_eh_frame_hdr failed");
    munmap(bytes, size);
    return -UNW_ENOINFO;
  }

  unw_addr_space_t image_as = unw_create_addr_space(&image_accessors, 0);
  unw_set_caching_policy(image_as, UNW_CACHE_NONE);
  int ret = unw_get_unwind_table((unw_word_t) ip, pi,
				 (int) needUnwindInfo,
				 peh_vaddr,
				 image_as, image);
  logf(env, fine, "post unw_get_unwind_table %d", ret);
  return ret;
}

jint
TARGET::fillProcInfoFromVDSO(jnixx::env env, jlong unwProcInfo, jlong ip,
			     bool needUnwindInfo,
			     AddressSpace addressSpace,
			     jlong lowAddress, jlong highAddress,
			     jlong offset) {
  logf(FINE, "fillProcInfoFromVDSO segbase: %llx segbound: %llx segoffset: %llx",
       (long long) lowAddress, (long long) highAddress, (long long) offset);
    void *image;
  unw_word_t magic;
  unw_accessors_t *a;

  unw_word_t segbase = (unw_word_t) lowAddress;
  unw_word_t segbound = (unw_word_t) highAddress;
  unw_word_t segoffset = (unw_word_t) offset;
  unw_word_t size = (unw_word_t) (segbound - segbase);
  if (size > MAX_VDSO_SIZE)
    return -UNW_ENOINFO;
  logf(FINE, "checked size, 0x%llx", (long long) size);

  logf(FINE, "checking access_mem");
  unw_addr_space_t as = (unw_addr_space_t) addressSpace.GetUnwAddressSpace(env);
  a = unw_get_accessors (as);
  if (! a->access_mem)
    return -UNW_ENOINFO;

  // Try to decide whether it's an ELF image before bringing it all in
  // from target memory.
  logf(FINE, "checking magic");
  if (size <= EI_CLASS || size <= sizeof (magic))
    return -UNW_ENOINFO;
  if (sizeof (magic) >= SELFMAG) {
    // Read the VDSO from the target.
    int ret = (*a->access_mem) (as, segbase, &magic,
				0, (void *) addressSpace._object);
    if (ret < 0) {
      logf(FINE, "error accessing VDSO %d", ret);
      return ret;
    }
    if (memcmp (&magic, ELFMAG, SELFMAG) != 0) {
      logf(FINE, "VDSO has bad magic");
      return -UNW_ENOINFO;
    }
  }

  logf(FINE, "mapping memory for image (using mmap, so can umaped)");
  image = mmap (0, size, PROT_READ | PROT_WRITE,
                MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
  if (image == MAP_FAILED)
    return -UNW_ENOINFO;

  logf(FINE, "checked magic");
  unw_word_t hi;
  if (sizeof (magic) >= SELFMAG) {
    *(unw_word_t *)image = magic;
    hi = sizeof (magic);
  } else {
    hi = 0;
  }

  logf(FINE, "reading in VDSO");
  for (; hi < size; hi += sizeof (unw_word_t)) {
    logf(env, GetFinest(env),
	 "Reading memory segbase: 0x%llx, image: %p, hi: 0x%llx, at: 0x%llx to location: %p",
	 (long long) segbase, image, (long long) hi, (long long) segbase + hi, (char *) image + hi);
    int ret = (*a->access_mem) (as, segbase + hi,
				(unw_word_t *) ((char *) image + hi),
				0, (void *) addressSpace._object);
    if (ret < 0) {
      logf(FINE, "error reading vdso");
      munmap (image, size);
      return ret;
    }
  }

  return fillProcInfoFromImage(FINE, "[vdso]", UNW_PROC_INFO,
			       needUnwindInfo, image, size,
			       ip, segbase, segbound, segoffset);
}

jint
TARGET::fillProcInfoFromElfImage(jnixx::env env, jlong unwProcInfo, jlong ip,
				 bool needUnwindInfo,
				 AddressSpace addressSpace,
				 String elfImageName,
				 jlong segbase, jlong segbound,
				 jlong segoffset) {
  logf(FINE, "fillProcInfoFromElfImage");
  struct stat stat;
  void *image;		/* pointer to mmap'd image */
  size_t size;		/* (file-) size of the image */
  jstringUTFChars name = jstringUTFChars(env, elfImageName);

  logf(FINE, "opening %s", name.elements());
  int fd = ::open(name.elements(), O_RDONLY);
  if (fd < 0) {
    int err = errno;
    logf(FINE, "open failed: %s", strerror(err));
    return -UNW_ENOINFO;
  }

  logf(FINE, "stat-ing %d", fd);
  int ret = ::fstat(fd, &stat);
  if (ret < 0) {
    int err = errno;
    ::close(fd);
    logf(FINE, "fstat failed: %s", strerror(err));
    return -UNW_ENOINFO;
  }

  size = stat.st_size;
  logf(FINE, "mmaping %d, size %ld", fd, (long) size);
  image = ::mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
  if (image == MAP_FAILED) {
    int err = errno;
    ::close(fd);
    logf(FINE, "mmap failed: %s", strerror(err));
    return -UNW_ENOINFO;
  }
  ::close(fd);
  
  return fillProcInfoFromImage(FINE, name.elements(), UNW_PROC_INFO,
			       needUnwindInfo, image, size,
			       ip, segbase, segbound, segoffset);
}

jint
TARGET::fillProcInfoNotAvailable(jnixx::env env, jlong unwProcInfo) {
  return -UNW_ENOINFO;
}

jlong
TARGET::getStartIP(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->start_ip;
}

jlong
TARGET::getEndIP(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->end_ip;
}

jlong
TARGET::getLSDA(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->lsda;
}

jlong
TARGET::getHandler(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->handler;
}

jlong
TARGET::getGP(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->gp;
}

jlong
TARGET::getFlags(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->flags;
}

jint
TARGET::getFormat(jnixx::env env, jlong unwProcInfo) {
  return (jint) (UNW_PROC_INFO)->format;
}

jint
TARGET::getUnwindInfoSize(jnixx::env env, jlong unwProcInfo) {
  return (jint) (UNW_PROC_INFO)->unwind_info_size;
}

jlong
TARGET::getUnwindInfo(jnixx::env env, jlong unwProcInfo) {
  return (jlong) (UNW_PROC_INFO)->unwind_info;
}
